/****************************************************************************
 *
 * $Id: //depot/prod/champagne/Rchampagne5.0.1x/src/rrdtool-1.0.28/src/ad_graph.c#1 $
 *
 *  Based on rrd_graph.c:
 *
 * RRDtool 1.0.28  Copyright Tobias Oetiker, 1997 - 2000
 *
 * Copyright 2003 NetApp, contribution by Brian Hackworth
 *
 ****************************************************************************/

#include "rrd_tool.h"

#include <gd.h>
#include <gdlucidan10.h>
#include <gdfontmb.h>
#include <sys/stat.h>

#include "ad_graph.h"

#ifdef WIN32
#define	strcasecmp	stricmp
#endif

#define	AD_GRAPH_FORMATTED_YLABEL_WIDTH	7
#define	AD_GRAPH_RIGHT_MARGIN_PIXELS	10
#define	AD_GRAPH_LEFT_MARGIN_PIXELS	0
#define	AD_GRAPH_TOP_MARGIN_PIXELS	4
#define	AD_GRAPH_BOTTOM_MARGIN_PIXELS	8
#define	AD_GRAPH_TITLE_PADDING_PIXELS	4
#define	AD_GRAPH_LEGEND_PADDING_PIXELS	4

/*
 *  Graphs smaller than this don't have any labels on them.
 */
#define	AD_GRAPH_TINY_THRESHOLD		50

/*
 *  Graphs taller than this have legends on them.
 */
#define	AD_GRAPH_LARGE_THRESHOLD	240

#define SmallFont gdLucidaNormal10
#define LargeFont gdFontMediumBold

#undef AD_GRAPH_DEBUG

#define	AD_GRAPH_LINE_LEN		1024
	/*
	 *  Arbitrary maximum length of lines of input data.
	 */

enum tmt_en {TMT_SECOND=0,TMT_MINUTE,TMT_HOUR,TMT_DAY,
	     TMT_WEEK,TMT_MONTH,TMT_YEAR};

enum grc_en {GRC_CANVAS=0,GRC_BACK,GRC_SHADEA,GRC_SHADEB,GRC_SHADEC,
	     GRC_GRID,GRC_MGRID,GRC_FONT,GRC_FRAME,GRC_ARROW,__GRC_END__};


typedef struct col_trip_t {
    int red; /* red = -1 is no color */
    int green;
    int blue;
    int i; /* color index assigned in gif image i=-1 is unasigned*/
} col_trip_t;


typedef struct xlab_t {
    long         minsec;       /* minimum sec per pix */
    enum tmt_en  gridtm;       /* grid interval in what ?*/
    long         gridst;       /* how many whats per grid*/
    enum tmt_en  mgridtm;      /* label interval in what ?*/
    long         mgridst;      /* how many whats per label*/
    enum tmt_en  labtm;        /* label interval in what ?*/
    long         labst;        /* how many whats per label*/
    long         precis;       /* label precision -> label placement*/
    char         *stst;        /* strftime string*/
} xlab_t;

xlab_t xlab[] = {
    {0,        TMT_SECOND,30, TMT_MINUTE,5,  TMT_MINUTE,5,         0,"%H:%M"},
    {2,        TMT_MINUTE,1,  TMT_MINUTE,5,  TMT_MINUTE,5,         0,"%H:%M"},
    {5,        TMT_MINUTE,2,  TMT_MINUTE,10, TMT_MINUTE,10,        0,"%H:%M"},
    {10,       TMT_MINUTE,5,  TMT_MINUTE,20, TMT_MINUTE,20,        0,"%H:%M"},
    {30,       TMT_MINUTE,10, TMT_HOUR,1,    TMT_HOUR,1,           0,"%H:%M"},
    {60,       TMT_MINUTE,30, TMT_HOUR,2,    TMT_HOUR,2,           0,"%H:%M"},
    {180,      TMT_HOUR,1,    TMT_HOUR,6,    TMT_HOUR,6,           0,"%H:%M"},
    /*{300,      TMT_HOUR,3,    TMT_HOUR,12,   TMT_HOUR,12,    12*3600,"%a %p"},  this looks silly*/
    {600,      TMT_HOUR,6,    TMT_DAY,1,     TMT_DAY,1,      24*3600,"%a"},
    {1800,     TMT_HOUR,12,   TMT_DAY,1,     TMT_DAY,2,      24*3600,"%a"},
    /* 
     *  NetApp: put Month and Day on the weekly graphs rather than week #
     */
    /*  This one is used for the small 1w graphs.  */
    {3600,     TMT_DAY,1,     TMT_DAY,1,    TMT_DAY,3,    0,"%d %b"},
    /*  This one is used for the large 1m graphs.  */
    {5400,   TMT_DAY,1,     TMT_DAY,7,    TMT_DAY,7,    0,"%d %b"},
    /*  This one is used for the small 1m graphs.  */
    {3*3600,   TMT_WEEK,1,    TMT_MONTH,1,   TMT_WEEK,2,    7*24*3600,"%d %b"},
    {6*3600,   TMT_MONTH,1,   TMT_MONTH,1,   TMT_MONTH,1, 30*24*3600,"%b"},
    {18*3600,  TMT_MONTH,1,   TMT_MONTH,3,   TMT_MONTH,3, 30*24*3600,"%b"},
    {10*24*3600, TMT_YEAR,1,  TMT_YEAR,1,    TMT_YEAR,1, 365*24*3600,"%y"},
    {-1,TMT_MONTH,0,TMT_MONTH,0,TMT_MONTH,0,0,""}
};

/* sensible logarithmic y label intervals ...
   the first element of each row defines the possible starting points on the
   y axis ... the other specify the */

double yloglab[][12]= {{ 1e9, 1,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0 },
		       {  1e3, 1,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0 },
		       {  1e1, 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
		       /* {  1e1, 1,  5,  0,  0,  0,  0,  0,  0,  0,  0,  0 }, */
		       {  1e1, 1,  2.5,  5,  7.5,  0,  0,  0,  0,  0,  0,  0 },
		       {  1e1, 1,  2,  4,  6,  8,  0,  0,  0,  0,  0,  0 },
		       {  1e1, 1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0 },
		       {  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 }};

/* sensible y label intervals ...*/

typedef struct ylab_t {
    double   grid;    /* grid spacing */
    int      lfac[4]; /* associated label spacing*/
} ylab_t;

ylab_t ylab[]= {
    {0.1, {1,2, 5,10}},
    {0.2, {1,5,10,20}},
    {0.5, {1,2, 4,10}},
    {1.0,   {1,2, 5,10}},
    {2.0,   {1,5,10,20}},
    {5.0,   {1,2, 4,10}},
    {10.0,  {1,2, 5,10}},
    {20.0,  {1,5,10,20}},
    {50.0,  {1,2, 4,10}},
    {100.0, {1,2, 5,10}},
    {200.0, {1,5,10,20}},
    {500.0, {1,2, 4,10}},
    {0.0,   {0,0,0,0}}};


col_trip_t graph_col[] = { /* default colors */
    {255,255,255,-1},   /* canvas */
    {231,232,235,-1},   /* background */
    {200,200,200,-1},   /* shade A    */
    {150,150,150,-1},   /* shade B    */
    {238,238,238,-1},   /* shade C; matches NetCharts    */
    {140,140,140,-1},      /* grid */
    {130,30,30,-1},      /* major grid */
    {17,21,53,-1},         /* font */ 	
    {17,21,53,-1},         /* frame */
    {255,0,0,-1}	/*arrow*/
};

/* this structure describes the elements which can make up a graph.
   because they are quite diverse, not all elements will use all the
   possible parts of the structure. */
#ifdef HAVE_SNPRINTF
#define FMT_LEG_LEN 200
#else
#define FMT_LEG_LEN 2000
#endif

typedef  struct graph_desc_t {
    ad_set_t *	   gd_set;
    col_trip_t     col;        /* graph color */
    char           format[FMT_LEG_LEN+5]; /* format for PRINT AND GPRINT */
    gdPoint        legloc;     /* location of legend */   
    double         yrule;      /* value for y rule line */
    time_t         xrule;      /* value for x rule line */

} graph_desc_t;

#define ALTYGRID          0x01  /* use alternative y grid algorithm */
#define ALTAUTOSCALE      0x02  /* use alternative algorithm to find lower and upper bounds */
#define ALTAUTOSCALE_MAX  0x04  /* use alternative algorithm to find upper bounds */

typedef struct image_desc_t {

    /* configuration of graph */

    ad_graph_opt_t	* opts;

    long           xsize,ysize;    /* graph area size in pixels */
    col_trip_t     graph_col[__GRC_END__]; /* real colors for the graph */   
    xlab_t         xlab_user;      /* user defined labeling for xaxis */
    char           xlab_form[200]; /* format for the label on the xaxis */

    double         ygridstep;      /* user defined step for y grid */
    int            ylabfact;       /* every how many y grid shall a label be written ? */

    int            rigid;          /* do not expand range even with 
				      values outside */
    int            logarithmic;    /* scale the yaxis logarithmic */
    
    /* status information */
    	    
    long           xorigin,yorigin;/* where is (0,0) of the graph */
    long           xgif,ygif;      /* total size of the gif */
    int		   tiny;	   /* non-zero if this is a tiny graph */
    int		   large;	   /* non-zero if this is a large graph */
    double         magfact;        /* numerical magnitude*/
    char           symbol;         /* magnitude symbol for y-axis */
    /* data elements */

    long  prt_c;                  /* number of print elements */
    long  gdes_c;                  /* number of graphics elements */
    graph_desc_t   *gdes;          /* points to an array of graph elements */

} image_desc_t;

/* Prototypes */
static int xtr(image_desc_t *,time_t);
static int ytr(image_desc_t *, double);
static void si_unit( image_desc_t *);
static void expand_range(image_desc_t *);
#ifdef ORIG
void reduce_data( enum cf_en,  unsigned long,  time_t *, time_t *,  unsigned long *,  unsigned long *,  rrd_value_t **);
#endif
static int data_proc( image_desc_t *);
static time_t find_first_time( time_t,  enum tmt_en,  long);
static time_t find_next_time( time_t,  enum tmt_en,  long);
static int leg_place(image_desc_t *);
static int horizontal_grid(gdImagePtr, image_desc_t *);
static int horizontal_log_grid(gdImagePtr, image_desc_t *);
static void vertical_grid( gdImagePtr, image_desc_t *);
static void axis_paint( image_desc_t *, gdImagePtr);
static void grid_paint( image_desc_t *, gdImagePtr);
static void data_paint( image_desc_t *, gdImagePtr);
static gdImagePtr MkLineBrush(image_desc_t *,long, int width);
static int graph_paint(image_desc_t *);

static void
ad_graph_string_to_color(const char * colstr, col_trip_t * col)
{
	if (0 == colstr || 0 == *colstr) {
		col->red = 255;
		col->green = 0;
		col->blue = 0;
	}
	else {
		sscanf(colstr, "#%2x%2x%2x", &col->red, &col->green, &col->blue);
	}
	col->i = -1;
}
/* translate time values into x coordinates */   
/*#define xtr(x) (int)((double)im->xorigin \
		+ ((double) im->xsize / (double)(im->opts->go_end - im->opts->go_start) ) \
		* ((double)(x) - im->opts->go_start)+0.5) */
/* initialize with xtr(im,0); */
static int
xtr(image_desc_t *im,time_t mytime){
    static double pixie;
    if (mytime==0){
	pixie = (double) im->xsize /
		(double)(im->opts->go_end - im->opts->go_start);
	return im->xorigin;
    }
    return (int)((double)im->xorigin 
		 + pixie * ( mytime - im->opts->go_start ) );
}

/* translate data values into y coordinates */

/* #define ytr(x) (int)((double)im->yorigin \
		- ((double) im->ysize / (im->opts->go_maxval - im->opts->go_minval) ) \
		* ((double)(x) - im->opts->go_minval)+0.5) */
static int
ytr(image_desc_t *im, double value){
    static double pixie;
    double yval;
    if (isnan(value)){
      if(!im->logarithmic)
	pixie = (double) im->ysize / (im->opts->go_maxval - im->opts->go_minval);
      else 
	pixie = (double) im->ysize / (log10(im->opts->go_maxval) - log10(im->opts->go_minval));
      yval = im->yorigin;
    } else if(!im->logarithmic) {
      yval = im->yorigin - pixie * (value - im->opts->go_minval) + 0.5;
    } else {
      if (value < im->opts->go_minval) {
	yval = im->yorigin;
      } else {
	yval = im->yorigin - pixie * (log10(value) - log10(im->opts->go_minval)) + 0.5;
      }
    }
    /* make sure we don't return anything too unreasonable. GD lib can
       get terribly slow when drawing lines outside its scope. This is 
       especially problematic in connection with the rigid option */
    if (! im->rigid) {
      return (int)yval;
    } else if ((int)yval > im->yorigin) {
      return im->yorigin+2;
    } else if ((int) yval < im->yorigin - im->ysize){
      return im->yorigin - im->ysize - 2;
    } else {
      return (int)yval;
    } 
}


/* find SI magnitude symbol for the numbers on the y-axis*/
static void 
si_unit(
    image_desc_t *im   /* image description */
)
{

    char symbol[] = {
		     ' ', /* Base */
		     'K', /* 10e3   Kilo */
		     'M', /* 10e6   Mega */
		     'G', /* 10e9   Giga */
		     'T', /* 10e12  Terra */
		     'P', /* 10e15  Peta */
		     'E'};/* 10e18  Exa */

    int   symbcenter = 0;
    double digits;  
    
    digits = floor( log( max( fabs(im->opts->go_minval),fabs(im->opts->go_maxval)))/log((double)im->opts->go_base)); 
    im->magfact = pow((double)im->opts->go_base , digits);
#ifdef AD_GRAPH_DEBUG
    fprintf(stderr, "digits %6.3f  im->magfact %6.3f\n",digits,im->magfact);
#endif
    if ( ((digits+symbcenter) < sizeof(symbol)) &&
		    ((digits+symbcenter) >= 0) )
        im->symbol = symbol[(int)digits+symbcenter];
    else {
	im->magfact = 1.0;
	im->symbol = ' ';
    }
 }

/*  move min and max values around to become sensible */

static void 
expand_range(image_desc_t *im)
{
	/*
 	 *  DFM-MOD
 	 *
 	 *  Grapher doesn't know how to pick sensible Y axis values
	 *  when the scaled value is between 1000 and 1024. Adding
	 *  1100.0 and 1050.0 to the start of sensiblevalues array.
 	 *
 	 *  See burt147240.
 	 */
    double sensiblevalues[] ={
			      1100.0, 1050.0,
			      1000.0,950.0,900.0,850.0,800.0,750.0,700.0,
			      650.0,600.0,550.0,500.0,450.0,400.0,
			      350.0,300.0,250.0,
			      200.0,150.0,125.0,100.0,95.0,90.0,80.0,
			      75.0,70.0,60.0,50.0,40.0,30.0,
			      25.0,20.0,10.0,9.0,8.0,
			      7.0,6.0,5.0,4.0,3.5,3.0,
			      2.5,2.0,1.8,1.5,1.2,1.0,
			      0.8,0.7,0.6,0.5,0.4,0.3,0.2,0.1,
			      0.00000001,-1};
    
    double scaled_min,scaled_max;  
    double adj;
    int i;
    

    
#ifdef AD_GRAPH_DEBUG
    fprintf(stderr, "Min: %6.5f Max: %6.5f MagFactor: %6.2f\n",
	   im->opts->go_minval,im->opts->go_maxval,im->magfact);
#endif

    if (isnan(im->ygridstep)){
	if(im->opts->go_extra_flags & ALTAUTOSCALE) {
	    /* measure the amplitude of the function. Make sure that
	       graph boundaries are slightly higher then max/min vals
	       so we can see amplitude on the graph */
	      adj = (im->opts->go_maxval - im->opts->go_minval) * 0.1;
	      im->opts->go_minval -= adj;
	      im->opts->go_maxval += adj;
	}
	else if(im->opts->go_extra_flags & ALTAUTOSCALE_MAX) {
	    /* measure the amplitude of the function. Make sure that
	       graph boundaries are slightly higher than max vals
	       so we can see amplitude on the graph */
	      adj = (im->opts->go_maxval - im->opts->go_minval) * 0.1;
	      im->opts->go_maxval += adj;
	}
	else {
	    scaled_min = (im->opts->go_minval / im->magfact);
	    scaled_max = (im->opts->go_maxval / im->magfact);
	    
	    for (i=1; sensiblevalues[i] > 0; i++){
		if (sensiblevalues[i-1]>=scaled_min &&
		    sensiblevalues[i]<=scaled_min)	
		    im->opts->go_minval = sensiblevalues[i]*(im->magfact);
		
		if (-sensiblevalues[i-1]<=scaled_min &&
		-sensiblevalues[i]>=scaled_min)
		    im->opts->go_minval = -sensiblevalues[i-1]*(im->magfact);
		
		if (sensiblevalues[i-1] >= scaled_max &&
		    sensiblevalues[i] <= scaled_max)
		    im->opts->go_maxval = sensiblevalues[i-1]*(im->magfact);
		
		if (-sensiblevalues[i-1]<=scaled_max &&
		    -sensiblevalues[i] >=scaled_max)
		    im->opts->go_maxval = -sensiblevalues[i]*(im->magfact);
	    }
	}
    } else {
	/* adjust min and max to the grid definition if there is one */
	im->opts->go_minval = (double)im->ylabfact * im->ygridstep * 
	    floor(im->opts->go_minval / ((double)im->ylabfact * im->ygridstep));
	im->opts->go_maxval = (double)im->ylabfact * im->ygridstep * 
	    ceil(im->opts->go_maxval /( (double)im->ylabfact * im->ygridstep));
    }
    
#ifdef AD_GRAPH_DEBUG
    fprintf(stderr,"SCALED Min: %6.5f Max: %6.5f Factor: %6.2f\n",
	   im->opts->go_minval,im->opts->go_maxval,im->magfact);
#endif
}

    

/* evaluate the expressions in the CDEF functions */

/*************************************************************
 * CDEF stuff 
 *************************************************************/

/* massage data so, that we get one value for each x coordinate in the graph */
static int
data_proc( image_desc_t *im ){
    size_t i,ii;

    ad_sample_t *	sample;
    double value;
    double minval=DNAN,maxval=DNAN;

    ad_set_t *	set;
    

    for ( i = 0; i < im->opts->go_set_count; i++ ) {

	set = im->opts->go_sets[i];
	for ( ii = 0; ii < set->ads_sample_count; ii++ ) {

		sample = set->ads_samples[ii];
		value = sample->s_value;

		if (!isnan(value)) {
			if (isnan(minval) || value < minval) {
				minval = value;
			}
			if (isnan(maxval) || value > maxval) {
				maxval = value;
			}
		}
	}
    }
    
    /* if min or max have not been asigned a value this is because
       there was no data in the graph ... this is not good ...
       lets set these to dummy values then ... */

    if (isnan(minval)) minval = 0.0;
    if (isnan(maxval)) maxval = 1.0;
    
    /* adjust min and max values */
    if (isnan(im->opts->go_minval) 
	|| ((!im->logarithmic && !im->rigid) /* don't adjust low-end with log scale */
	    && im->opts->go_minval > minval))
	im->opts->go_minval = minval;
    if (isnan(im->opts->go_maxval) 
	|| (!im->rigid 
	    && im->opts->go_maxval < maxval)){
	if (im->logarithmic)
	    im->opts->go_maxval = maxval * 1.1;
	else
	    im->opts->go_maxval = maxval;
    }
    /* make sure min and max are not equal */
    if (im->opts->go_minval == im->opts->go_maxval) {
      im->opts->go_maxval *= 1.01; 
      if (! im->logarithmic) {
	im->opts->go_minval *= 0.99;
      }
      
      /* make sure min and max are not both zero */
      if (im->opts->go_maxval == 0.0) {
	    im->opts->go_maxval = 1.0;
      }
        
    }
    return 0;
}



/* identify the point where the first gridline, label ... gets placed */

static time_t
find_first_time(
    time_t   start, /* what is the initial time */
    enum tmt_en baseint,  /* what is the basic interval */
    long     basestep /* how many if these do we jump a time */
    )
{
    struct tm tm;
    tm = *localtime(&start);
    switch(baseint){
    case TMT_SECOND:
	tm.tm_sec -= tm.tm_sec % basestep; break;
    case TMT_MINUTE: 
	tm.tm_sec=0;
	tm.tm_min -= tm.tm_min % basestep; 
	break;
    case TMT_HOUR:
	tm.tm_sec=0;
	tm.tm_min = 0;
	tm.tm_hour -= tm.tm_hour % basestep; break;
    case TMT_DAY:
	/* we do NOT look at the basestep for this ... */
	tm.tm_sec=0;
	tm.tm_min = 0;
	tm.tm_hour = 0; break;
    case TMT_WEEK:
	/* we do NOT look at the basestep for this ... */
	tm.tm_sec=0;
	tm.tm_min = 0;
	tm.tm_hour = 0;
	tm.tm_mday -= tm.tm_wday -1;	/* -1 because we want the monday */
	if (tm.tm_wday==0) tm.tm_mday -= 7; /* we want the *previous* monday */
	break;
    case TMT_MONTH:
	tm.tm_sec=0;
	tm.tm_min = 0;
	tm.tm_hour = 0;
	tm.tm_mday = 1;
	tm.tm_mon -= tm.tm_mon % basestep; break;

    case TMT_YEAR:
	tm.tm_sec=0;
	tm.tm_min = 0;
	tm.tm_hour = 0;
	tm.tm_mday = 1;
	tm.tm_mon = 0;
	tm.tm_year -= (tm.tm_year+1900) % basestep;
	
    }
    return mktime(&tm);
}
/* identify the point where the next gridline, label ... gets placed */
static time_t 
find_next_time(
    time_t   current, /* what is the initial time */
    enum tmt_en baseint,  /* what is the basic interval */
    long     basestep /* how many if these do we jump a time */
    )
{
    struct tm tm;
    time_t madetime;
    tm = *localtime(&current);
    do {
	switch(baseint){
	case TMT_SECOND:
	    tm.tm_sec += basestep; break;
	case TMT_MINUTE: 
	    tm.tm_min += basestep; break;
	case TMT_HOUR:
	    tm.tm_hour += basestep; break;
	case TMT_DAY:
	    tm.tm_mday += basestep; break;
	case TMT_WEEK:
	    tm.tm_mday += 7*basestep; break;
	case TMT_MONTH:
	    tm.tm_mon += basestep; break;
	case TMT_YEAR:
	    tm.tm_year += basestep;	
	}
	madetime = mktime(&tm);
    } while (madetime == -1); /* this is necessary to skip impssible times
				 like the daylight saving time skips */
    return madetime;
	  
}



/* place legends with color spots */
static int
leg_place(image_desc_t *im)
{
    
    /* graph labels */
    int   interleg = SmallFont->w*2;
    int   box = (int) (SmallFont->h*1.2);
    int   border = SmallFont->w*2;
    int   fill=0, fill_last;
    int   leg_c = 0;
    int   leg_x = border, leg_y = im->yorigin;
    int   leg_cc;
    int   glue = 0;
    int   i,ii, mark = 0;
    char  prt_fctn; /*special printfunctions */
    int  *legspace;

    if (!im->opts->go_legend) {
	return 0;
    }
    if (im->tiny) {
	return 0;
    }
    if ((legspace = malloc(im->gdes_c*sizeof(int)))==NULL){
       rrd_set_error("malloc for legspace");
       return -1;
    }

    for(i=0;i<im->gdes_c;i++){
	fill_last = fill;

	leg_cc = strlen(im->gdes[i].gd_set->ads_name);
	
	/* is there a controle code ant the end of the legend string ? */ 
	if (leg_cc >= 2 && im->gdes[i].gd_set->ads_name[leg_cc-2] == '\\') {
	    prt_fctn = im->gdes[i].gd_set->ads_name[leg_cc-1];
	    leg_cc -= 2;
	    im->gdes[i].gd_set->ads_name[leg_cc] = '\0';
	} else {
	    prt_fctn = '\0';
	}
        /* remove exess space */
        while (prt_fctn=='g' && 
	       leg_cc > 0 && 
	       im->gdes[i].gd_set->ads_name[leg_cc-1]==' '){
	   leg_cc--;
	   im->gdes[i].gd_set->ads_name[leg_cc]='\0';
	}
	if (leg_cc != 0 ){	    
	   legspace[i]=(prt_fctn=='g' ? 0 : interleg);
	   
	   if (fill > 0){ 
 	       /* no interleg space if string ends in \g */
	       fill += legspace[i];
	    }
	    if (1) {
		fill += box; 	   
	    }
	    fill += leg_cc * SmallFont->w;
	    leg_c++;
	} else {
	   legspace[i]=0;
	}
        /* who said there was a special tag ... ?*/
	if (prt_fctn=='g') {    
	   prt_fctn = '\0';
	}
	if (prt_fctn == '\0') {
	    if (i == im->gdes_c -1 ) prt_fctn ='l';
	    
	    /* is it time to place the legends ? */
	    if (fill > im->xgif - 2*border){
		if (leg_c > 1) {
		    /* go back one */
		    i--; 
		    fill = fill_last;
		    leg_c--;
		    prt_fctn = 'j';
		} else {
		    prt_fctn = 'l';
		}
		
	    }
	}


	if (prt_fctn != '\0'){
	    leg_x = border;
	    if (leg_c >= 2 && prt_fctn == 'j') {
		glue = (im->xgif - fill - 2* border) / (leg_c-1);
		/* if (glue > 2 * SmallFont->w) glue = 0; */
	    } else {
		glue = 0;
	    }
	    if (prt_fctn =='c') leg_x =  (im->xgif - fill) / 2;
	    if (prt_fctn =='r') leg_x =  im->xgif - fill - border;

	    for(ii=mark;ii<=i;ii++){
		if(im->gdes[ii].gd_set->ads_name[0]=='\0')
		    continue;
		im->gdes[ii].legloc.x = leg_x;
		im->gdes[ii].legloc.y = leg_y;
		leg_x =  leg_x 
		    + strlen(im->gdes[ii].gd_set->ads_name)*SmallFont->w 
		    + legspace[ii]
		    + glue;
		if (1) {
		    leg_x += box; 	   
		}
	    }	    
	    leg_y = leg_y + (int) (SmallFont->h*1.2);
	    if (prt_fctn == 's') leg_y -= (int) (SmallFont->h *0.5);
	    fill = 0;
	    leg_c = 0;
	    mark = ii;
	}	   
    }
    im->ysize -= (box + AD_GRAPH_LEGEND_PADDING_PIXELS);
    im->yorigin -= (box + AD_GRAPH_LEGEND_PADDING_PIXELS);
    free(legspace);
    return 0;
}


static int
horizontal_grid(gdImagePtr gif, image_desc_t   *im)
{
    double   range;
    double   scaledrange;
    int      pixel,i;
    int      sgrid,egrid;
    double   gridstep;
    double   scaledstep;
    char     graph_label[100];
    gdPoint  polyPoints[4];
    int      labfact,gridind;
    int      styleMinor[2],styleMajor[2];
    int      decimals, fractionals;
    char     labfmt[64];
    int      bands = -1;

    labfact=2;
    gridind=-1;
    range =  im->opts->go_maxval - im->opts->go_minval;
    scaledrange = range / im->magfact;

	/* does the scale of this graph make it impossible to put lines
	   on it? If so, give up. */
	if (isnan(scaledrange)) {
		return 0;
	}

    styleMinor[0] = graph_col[GRC_GRID].i;
    styleMinor[1] = gdTransparent;

    styleMajor[0] = graph_col[GRC_MGRID].i;
    styleMajor[1] = gdTransparent;

    /* find grid spaceing */
    pixel=1;
    if(isnan(im->ygridstep)){
	if(im->opts->go_extra_flags & ALTYGRID) {
	    /* find the value with max number of digits. Get number of digits */
	    decimals = (int) ceil(log10(max(fabs(im->opts->go_maxval), fabs(im->opts->go_minval))));
	    if(decimals <= 0) /* everything is small. make place for zero */
		decimals = 1;
	    
	    fractionals = (int) floor(log10(range));
	    if(fractionals < 0) /* small amplitude. */
		sprintf(labfmt, "%%%d.%df", decimals - fractionals + 1, -fractionals + 1);
	    else
		sprintf(labfmt, "%%%d.1f", decimals + 1);
	    gridstep = pow((double)10, (double)fractionals);
	    if(gridstep == 0) /* range is one -> 0.1 is reasonable scale */
		gridstep = 0.1;
	    /* should have at least 5 lines but no more then 15 */
	    if(range/gridstep < 5)
                gridstep /= 10;
	    if(range/gridstep > 15)
                gridstep *= 10;
	    if(range/gridstep > 5) {
		labfact = 1;
		if(range/gridstep > 8)
		    labfact = 2;
	    }
	    else {
		gridstep /= 5;
		labfact = 5;
	    }
	}
	else {
	    for(i=0;ylab[i].grid > 0;i++){
		pixel = (int) ((double) im->ysize / (scaledrange / ylab[i].grid));
		if (gridind == -1 && pixel > 5) {
		    gridind = i;
		    break;
		}
	    }

	    if (-1 == gridind) {
		gridind = 0;
	    }
	    
	    for(i=0; i<4;i++) {
		if (pixel * ylab[gridind].lfac[i] >=  2 * SmallFont->h) {
		    labfact =  ylab[gridind].lfac[i];
		    break;
		}		          
	    } 
	    
	    gridstep = ylab[gridind].grid * im->magfact;
	}
    } else {
	gridstep = im->ygridstep;
	labfact = im->ylabfact;
    }
    
    polyPoints[0].x=im->xorigin;
    polyPoints[1].x=im->xorigin+im->xsize;
    polyPoints[1].y=-1;
    sgrid = (int)( im->opts->go_minval / gridstep - 1);
    egrid = (int)( im->opts->go_maxval / gridstep + 1);
    scaledstep = gridstep/im->magfact;

    for (i = sgrid; i <= egrid; i++){
	polyPoints[0].y=ytr(im,gridstep*i);
	if ( polyPoints[0].y >= im->yorigin-im->ysize
	     && polyPoints[0].y <= im->yorigin) {
	    if (bands < 0) {
		bands = 1;
	    }
	    if(i % labfact == 0){		
		if (i==0 || im->symbol == ' ') {
		    if(scaledstep < 1){
			if(im->opts->go_extra_flags & ALTYGRID) {
			    sprintf(graph_label,labfmt,scaledstep*i);
			}
			else {
			    sprintf(graph_label,"%4.1f %s",scaledstep*i,
					im->opts->go_ysuffix);
			}
		    } else {
			sprintf(graph_label,"%4.0f %s",scaledstep*i,
					im->opts->go_ysuffix);
		    }
		}else {
		    if(scaledstep < 1){
			sprintf(graph_label,"%4.1f %c%s",
				scaledstep*i, im->symbol,
				im->opts->go_ysuffix);
		    } else {
			sprintf(graph_label,"%4.0f %c%s",
				scaledstep*i, im->symbol,
				im->opts->go_ysuffix);
		    }
		}


		if (!im->tiny) {
			gdImageString(gif, SmallFont,
                              (polyPoints[0].x - (strlen(graph_label) * 
                                                  SmallFont->w)-7), 
                              polyPoints[0].y - SmallFont->h/2+1,
                              (unsigned char *)graph_label, graph_col[GRC_FONT].i);
		}
		
		gdImageSetStyle(gif, styleMajor, 2);

		gdImageLine(gif, polyPoints[0].x-2,polyPoints[0].y,
			    polyPoints[0].x+2,polyPoints[0].y,graph_col[GRC_MGRID].i);
		bands++;

		if (im->opts->go_hgrid & AD_GRID_MAJOR) {
			if (!(im->opts->go_hgrid & AD_GRID_BAND)) {
			gdImageLine(gif, polyPoints[1].x-2,polyPoints[0].y,
			    polyPoints[1].x+2,polyPoints[0].y,graph_col[GRC_MGRID].i);		    
		        gdImageLine(gif, polyPoints[0].x,polyPoints[0].y,
			    polyPoints[1].x,polyPoints[0].y,gdStyled);
			}
			else if (i != 0) {

		            if ((bands % 2) == 0) {
			        gdImageFilledRectangle(gif,
					polyPoints[0].x+1, polyPoints[0].y+1,
					polyPoints[1].x-1, polyPoints[1].y,
					graph_col[GRC_SHADEC].i);
		            }
			    if (polyPoints[0].y == im->yorigin) {
				gdImageLine(gif, polyPoints[1].x-2,polyPoints[0].y,
				    polyPoints[1].x+2,polyPoints[0].y,graph_col[GRC_MGRID].i);		    
				gdImageLine(gif, polyPoints[0].x,polyPoints[0].y,
				    polyPoints[1].x,polyPoints[0].y,gdStyled);
			    }

			}
		    polyPoints[1].y = polyPoints[0].y;
		}
	    } else {		
		gdImageSetStyle(gif, styleMinor, 2);
		gdImageLine(gif, polyPoints[0].x-1,polyPoints[0].y,
			    polyPoints[0].x+1,polyPoints[0].y,graph_col[GRC_GRID].i);
		if (im->opts->go_hgrid & AD_GRID_MINOR) {

			gdImageLine(gif, polyPoints[1].x-1,polyPoints[0].y,
			    polyPoints[1].x+1,polyPoints[0].y,graph_col[GRC_GRID].i);		    
		    gdImageLine(gif, polyPoints[0].x,polyPoints[0].y,
			polyPoints[1].x,polyPoints[0].y,gdStyled);
		}
	    }	    
	}	
    } 
    /*
     *  Finish up any band, if necessary.
     */
    if (im->opts->go_hgrid & AD_GRID_BAND) {
	    bands++;
	    if ((bands % 2) == 0) {
		gdImageFilledRectangle(gif,
			polyPoints[0].x+1, im->yorigin-im->ysize+1,
			polyPoints[1].x-1, polyPoints[1].y,
			graph_col[GRC_SHADEC].i);
	    }
    }
/*    if(im->opts->go_minval * im->opts->go_maxval < 0){
      polyPoints[0].y=ytr(0);
      gdImageLine(gif, polyPoints[0].x,polyPoints[0].y,
      polyPoints[1].x,polyPoints[0].y,graph_col[GRC_MGRID].i);
      } */

    return 1;
}

/* logaritmic horizontal grid */
static int
horizontal_log_grid(gdImagePtr gif, image_desc_t   *im)
{
    double   pixpex;
    int      ii,i;
    int      minoridx=0, majoridx=0;
    char     graph_label[100];
    gdPoint  polyPoints[4];
    int      styleMinor[2],styleMajor[2];
    double   value, pixperstep, minstep;

    /* find grid spaceing */
    pixpex= (double)im->ysize / (log10(im->opts->go_maxval) - log10(im->opts->go_minval));

	if (isnan(pixpex)) {
		return 0;
	}

    for(i=0;yloglab[i][0] > 0;i++){
	minstep = log10(yloglab[i][0]);
	for(ii=1;yloglab[i][ii+1] > 0;ii++){
	    if(yloglab[i][ii+2]==0){
		minstep = log10(yloglab[i][ii+1])-log10(yloglab[i][ii]);
		break;
	    }
	}
	pixperstep = pixpex * minstep;
	if(pixperstep > 5){minoridx = i;}
	if(pixperstep > 2 * SmallFont->h){majoridx = i;}
    }
   
    styleMinor[0] = graph_col[GRC_GRID].i;
    styleMinor[1] = gdTransparent;

    styleMajor[0] = graph_col[GRC_MGRID].i;
    styleMajor[1] = gdTransparent;

    polyPoints[0].x=im->xorigin;
    polyPoints[1].x=im->xorigin+im->xsize;
    /* paint minor grid */
    for (value = pow((double)10, log10(im->opts->go_minval) 
			  - fmod(log10(im->opts->go_minval),log10(yloglab[minoridx][0])));
	 value  <= im->opts->go_maxval;
	 value *= yloglab[minoridx][0]){
	if (value < im->opts->go_minval) continue;
	i=0;	
	while(yloglab[minoridx][++i] > 0){	    
	    polyPoints[0].y = ytr(im,value * yloglab[minoridx][i]);
	    if (polyPoints[0].y <= im->yorigin - im->ysize) break;
	    gdImageSetStyle(gif, styleMinor, 2);
	    gdImageLine(gif, polyPoints[0].x-1,polyPoints[0].y,
			polyPoints[0].x+1,polyPoints[0].y,graph_col[GRC_GRID].i);
	    gdImageLine(gif, polyPoints[1].x-1,polyPoints[0].y,
			polyPoints[1].x+1,polyPoints[0].y,graph_col[GRC_GRID].i);	    

	    gdImageLine(gif, polyPoints[0].x,polyPoints[0].y,
			polyPoints[1].x,polyPoints[0].y,gdStyled);
	}
    }

    /* paint major grid and labels*/
    for (value = pow((double)10, log10(im->opts->go_minval) 
			  - fmod(log10(im->opts->go_minval),log10(yloglab[majoridx][0])));
	 value <= im->opts->go_maxval;
	 value *= yloglab[majoridx][0]){
	if (value < im->opts->go_minval) continue;
	i=0;	
	while(yloglab[majoridx][++i] > 0){	    
	    polyPoints[0].y = ytr(im,value * yloglab[majoridx][i]);	    
	    if (polyPoints[0].y <= im->yorigin - im->ysize) break;
	    gdImageSetStyle(gif, styleMajor, 2);
	    gdImageLine(gif, polyPoints[0].x-2,polyPoints[0].y,
			polyPoints[0].x+2,polyPoints[0].y,graph_col[GRC_MGRID].i);
	    gdImageLine(gif, polyPoints[1].x-2,polyPoints[0].y,
			polyPoints[1].x+2,polyPoints[0].y,graph_col[GRC_MGRID].i);		    
	    
	    gdImageLine(gif, polyPoints[0].x,polyPoints[0].y,
			polyPoints[1].x,polyPoints[0].y,gdStyled);
	    sprintf(graph_label,"%3.0e",value * yloglab[majoridx][i]);

	    if (!im->tiny) {
		    gdImageString(gif, SmallFont,
			  (polyPoints[0].x - (strlen(graph_label) * 
					      SmallFont->w)-7), 
			  polyPoints[0].y - SmallFont->h/2+1,
			  (unsigned char *)graph_label, graph_col[GRC_FONT].i);	
	    }
	} 
    }
	return 1;
}


static void
vertical_grid(
    gdImagePtr     gif,
    image_desc_t   *im )
{   
    int xlab_sel;		/* which sort of label and grid ? */
    time_t ti, tilab;
    long factor;
    char graph_label[100];
    gdPoint polyPoints[4];	 /* points for filled graph and more*/

    /* style for grid lines */
    int     styleDotted[4];

    
    /* the type of time grid is determined by finding
       the number of seconds per pixel in the graph */
    
    
    if(im->xlab_user.minsec == -1){
	factor=(im->opts->go_end - im->opts->go_start)/im->xsize;
	xlab_sel=0;
	while ( xlab[xlab_sel+1].minsec != -1 
		&& xlab[xlab_sel+1].minsec <= factor){ xlab_sel++; }
	im->xlab_user.gridtm = xlab[xlab_sel].gridtm;
	im->xlab_user.gridst = xlab[xlab_sel].gridst;
	im->xlab_user.mgridtm = xlab[xlab_sel].mgridtm;
	im->xlab_user.mgridst = xlab[xlab_sel].mgridst;
	im->xlab_user.labtm = xlab[xlab_sel].labtm;
	im->xlab_user.labst = xlab[xlab_sel].labst;
	im->xlab_user.precis = xlab[xlab_sel].precis;
	im->xlab_user.stst = xlab[xlab_sel].stst;
    }
    
    /* y coords are the same for every line ... */
    polyPoints[0].y = im->yorigin;
    polyPoints[1].y = im->yorigin-im->ysize;

    /* paint the minor grid */
    for(ti = find_first_time(im->opts->go_start,
			    im->xlab_user.gridtm,
			    im->xlab_user.gridst);
	ti < im->opts->go_end; 
	ti = find_next_time(ti,im->xlab_user.gridtm,im->xlab_user.gridst)
	){
	/* are we inside the graph ? */
	if (ti < im->opts->go_start || ti > im->opts->go_end) continue;
	polyPoints[0].x = xtr(im,ti);
	styleDotted[0] = graph_col[GRC_GRID].i;
	styleDotted[1] = gdTransparent;

	gdImageSetStyle(gif, styleDotted, 2);

	/*
	 *  Only draw the gridlines if asked to do so.
	 */
	if (im->opts->go_vgrid & AD_GRID_MINOR) {
		gdImageLine(gif, polyPoints[0].x,polyPoints[0].y,
		    polyPoints[0].x,polyPoints[1].y,gdStyled);
	}
	/*
	 *  Always do the bottom hash marks:
	 */
	gdImageLine(gif, polyPoints[0].x,polyPoints[0].y-1,
		    polyPoints[0].x,polyPoints[0].y+1,graph_col[GRC_GRID].i);
	/*
	 *  Only do the top hash marks if we're also drawing the grid:
	 */
	if (im->opts->go_vgrid & AD_GRID_MINOR) {
		gdImageLine(gif, polyPoints[0].x,polyPoints[1].y-1,
		    polyPoints[0].x,polyPoints[1].y+1,graph_col[GRC_GRID].i);
	}
    }

    /* paint the major grid */
    for(ti = find_first_time(im->opts->go_start,
			    im->xlab_user.mgridtm,
			    im->xlab_user.mgridst);
	ti < im->opts->go_end; 
	ti = find_next_time(ti,im->xlab_user.mgridtm,im->xlab_user.mgridst)
	){
	/* are we inside the graph ? */
	if (ti < im->opts->go_start || ti > im->opts->go_end) continue;
	polyPoints[0].x = xtr(im,ti);
	styleDotted[0] = graph_col[GRC_MGRID].i;
	styleDotted[1] = gdTransparent;
	gdImageSetStyle(gif, styleDotted, 2);

	/*
	 *  Only draw the gridlines if asked to do so.
	 */
	if (im->opts->go_vgrid & AD_GRID_MAJOR) {
		gdImageLine(gif, polyPoints[0].x,polyPoints[0].y,
		    polyPoints[0].x,polyPoints[1].y,gdStyled);
	}
	else if (im->opts->go_hgrid & AD_GRID_BAND) {
		/*
		 *  Erase a small portion of the horizontal grid
		 *  as a visual indicator of the vertical grid.
		 */
		gdImageLine(gif, polyPoints[0].x,polyPoints[0].y,
		    polyPoints[0].x,polyPoints[1].y+1,graph_col[GRC_CANVAS].i);
	}
	/*
	 *  Always do the bottom hash marks:
	 */
	gdImageLine(gif, polyPoints[0].x,polyPoints[0].y-2,
		    polyPoints[0].x,polyPoints[0].y+2,graph_col[GRC_MGRID].i);
	/*
	 *  Only do the top hash marks if we're also drawing the grid:
	 */
	if (im->opts->go_vgrid & AD_GRID_MAJOR) {
		gdImageLine(gif, polyPoints[0].x,polyPoints[1].y-2,
		    polyPoints[0].x,polyPoints[1].y+2,graph_col[GRC_MGRID].i);
	}
    }
    /* paint the labels below the graph */
    for(ti = find_first_time(im->opts->go_start,
			    im->xlab_user.labtm,
			    im->xlab_user.labst);
	ti <= im->opts->go_end; 
	ti = find_next_time(ti,im->xlab_user.labtm,im->xlab_user.labst)
	){
	int gr_pos,width;
        tilab= ti + im->xlab_user.precis/2; /* correct time for the label */

#if HAVE_STRFTIME
	strftime(graph_label,99,im->xlab_user.stst,localtime(&tilab));
#else
# error "your libc has no strftime I guess we'll abort the exercise here."
#endif
	width=strlen(graph_label) *  SmallFont->w;
	gr_pos=xtr(im,tilab) - width/2;
	if (gr_pos  >= im->xorigin 
	    && gr_pos + width <= im->xorigin+im->xsize) 
	    gdImageString(gif, SmallFont,
			  gr_pos,  polyPoints[0].y+4,
			  (unsigned char *)graph_label, graph_col[GRC_FONT].i);
    }

}


static void 
axis_paint(
    image_desc_t   *im,
    gdImagePtr     gif
    )
{   
    /* draw x and y axis */
    gdImageLine(gif, im->xorigin+im->xsize,im->yorigin,
		im->xorigin+im->xsize,im->yorigin-im->ysize,
		graph_col[GRC_GRID].i);
    
    gdImageLine(gif, im->xorigin,im->yorigin-im->ysize,
		im->xorigin+im->xsize,im->yorigin-im->ysize,
		graph_col[GRC_GRID].i);

    gdImageLine(gif, im->xorigin-4,im->yorigin,
		im->xorigin+im->xsize+4,im->yorigin,
		graph_col[GRC_FONT].i);

    gdImageLine(gif, im->xorigin,im->yorigin,
		im->xorigin,im->yorigin-im->ysize,
		graph_col[GRC_GRID].i);
    
    if (im->opts->go_arrow) {
    /* arrow for X axis direction */
    gdImageLine(gif, im->xorigin+im->xsize+4, im->yorigin-3, im->xorigin+im->xsize+4, im->yorigin+3,graph_col[GRC_ARROW].i);
    gdImageLine(gif, im->xorigin+im->xsize+4, im->yorigin-3, im->xorigin+im->xsize+9, im->yorigin,graph_col[GRC_ARROW].i);
    gdImageLine(gif, im->xorigin+im->xsize+4, im->yorigin+3, im->xorigin+im->xsize+9, im->yorigin,graph_col[GRC_ARROW].i);
    }

    /*    gdImageLine(gif, im->xorigin+im->xsize-1, im->yorigin-3, im->xorigin+im->xsize-1, im->yorigin+3,graph_col[GRC_MGRID].i);
    gdImageLine(gif, im->xorigin+im->xsize, im->yorigin-2, im->xorigin+im->xsize, im->yorigin+2,graph_col[GRC_MGRID].i);
    gdImageLine(gif, im->xorigin+im->xsize+1, im->yorigin-2, im->xorigin+im->xsize+1, im->yorigin+2,graph_col[GRC_MGRID].i);
    gdImageLine(gif, im->xorigin+im->xsize+2, im->yorigin-2, im->xorigin+im->xsize+2, im->yorigin+2,graph_col[GRC_MGRID].i);
    gdImageLine(gif, im->xorigin+im->xsize+3, im->yorigin-1, im->xorigin+im->xsize+3, im->yorigin+1,graph_col[GRC_MGRID].i);
    gdImageLine(gif, im->xorigin+im->xsize+4, im->yorigin-1, im->xorigin+im->xsize+4, im->yorigin+1,graph_col[GRC_MGRID].i);
    gdImageLine(gif, im->xorigin+im->xsize+5, im->yorigin, im->xorigin+im->xsize+5, im->yorigin,graph_col[GRC_MGRID].i); */



}

static void
data_paint(
    image_desc_t   *im,
    gdImagePtr     gif
    )
{   
    size_t	i;
    size_t ii;
    ad_set_t *	set;
  gdImagePtr brush = 0;
	ad_sample_t *	sample;
	double value;
	time_t x;

    for ( i = 0; i < im->opts->go_set_count; i++) {	

	unsigned long lastx = im->xorigin;
	unsigned long lasty = im->yorigin - im->ysize;
	int		drawn = 0;


	set = im->opts->go_sets[i];
	brush = MkLineBrush(im,i,set->ads_width);
	gdImageSetBrush(gif, brush);

	for ( ii = 0; ii < set->ads_sample_count; ii++ ) {

	       sample = set->ads_samples[ii];
	       x = sample->s_time;
	       value = sample->s_value;

#ifdef AD_GRAPH_DEBUG
fprintf(stderr, "Row %d, time = %d, x = %d, value = %f\n", ii, (int) x, xtr(im,x), value);
#endif
		switch (set->ads_style) {
			default:
			case Graph_Line:
				break;
			
			case Graph_Bar:
				lastx = xtr(im,x);
				lasty = ytr(im,im->opts->go_minval) - set->ads_width/2;
				drawn = 1;
				break;
			
			case Graph_Point:
				lastx = xtr(im,x);
				lasty = ytr(im,value);
				drawn = 1;
				break;
		}

	       if (isnan(value))
			continue;

	       if (drawn) {
			 gdImageLine(gif,
			    lastx, lasty,
			    xtr(im,x),
			    ytr(im,value),
			    gdBrushed);
	       }
	       else {
			 gdImageLine(gif,
			    xtr(im,x),
			    ytr(im,value),
			    xtr(im,x),
			    ytr(im,value),
			    gdBrushed);
	       }
			 drawn = 1;
		
			lastx = xtr(im,x);
			lasty = ytr(im,value);
	    }
	}
	gdImageDestroy(brush);
}


static void
grid_paint(
    image_desc_t   *im,
    gdImagePtr     gif
    )
{   
    long i;
    int boxH=8, boxV=8;
	int res;
    gdPoint polyPoints[4];	 /* points for filled graph and more*/

    /* draw 3d border */
    gdImageLine(gif,0,0,im->xgif-1,0,graph_col[GRC_SHADEA].i);
    gdImageLine(gif,1,1,im->xgif-2,1,graph_col[GRC_SHADEA].i);
    gdImageLine(gif,0,0,0,im->ygif-1,graph_col[GRC_SHADEA].i);
    gdImageLine(gif,1,1,1,im->ygif-2,graph_col[GRC_SHADEA].i);
    gdImageLine(gif,im->xgif-1,0,im->xgif-1,im->ygif-1,graph_col[GRC_SHADEB].i);
    gdImageLine(gif,0,im->ygif-1,im->xgif-1,im->ygif-1,graph_col[GRC_SHADEB].i);
    gdImageLine(gif,im->xgif-2,1,im->xgif-2,im->ygif-2,graph_col[GRC_SHADEB].i);
    gdImageLine(gif,1,im->ygif-2,im->xgif-2,im->ygif-2,graph_col[GRC_SHADEB].i);


    
    /* dont draw horizontal grid if there is no min and max val */
	if(im->logarithmic){
		res = horizontal_log_grid(gif,im);
	} else {
		res = horizontal_grid(gif,im);
	}

    vertical_grid(gif, im);
    if (! res) {
       char *nodata = "There is not enough data for this graph";
       gdImageString(gif, LargeFont,
		     im->xgif/2 
		     - (strlen(nodata)*LargeFont->w)/2,
		     (2*im->yorigin-im->ysize) / 2,
		     (unsigned char *)nodata, graph_col[GRC_FONT].i);
    }


#ifdef	AD_GRAPH_DRAW_YAXIS_LEGEND
    /* yaxis description */
    gdImageStringUp(gif, SmallFont,
		    7,
		    (im->yorigin - im->ysize/2
		     +(strlen(im->opts->go_ylegend)*SmallFont->w)/2 ),
		    (unsigned char *)im->opts->go_ylegend, graph_col[GRC_FONT].i);
    
#endif

    if (!im->tiny && 0 != im->opts->go_title[0]) {
    /* graph title */
    gdImageString(gif, LargeFont,
		    im->xgif/2 
		    - (strlen(im->opts->go_title)*LargeFont->w)/2,
		  AD_GRAPH_TOP_MARGIN_PIXELS,
		    (unsigned char *)im->opts->go_title, graph_col[GRC_FONT].i);
    
    }

    /* graph labels */

    if (im->opts->go_legend && !im->tiny) {
	for(i=0;i<im->gdes_c;i++){
		if(im->gdes[i].gd_set->ads_name[0] =='\0')
		    continue;

		if(1) {
		    
		    polyPoints[0].x = im->gdes[i].legloc.x;
		    polyPoints[0].y = im->gdes[i].legloc.y+1;
		    polyPoints[1].x = polyPoints[0].x+boxH;
		    polyPoints[2].x = polyPoints[0].x+boxH;
		    polyPoints[3].x = polyPoints[0].x;
		    polyPoints[1].y = polyPoints[0].y;
		    polyPoints[2].y = polyPoints[0].y+boxV;
		    polyPoints[3].y = polyPoints[0].y+boxV;
		    gdImageFilledPolygon(gif,polyPoints,4,im->gdes[i].col.i);
		    gdImagePolygon(gif,polyPoints,4,graph_col[GRC_FRAME].i);

		    gdImageString(gif, SmallFont,
				  polyPoints[0].x+boxH+6, 
				  polyPoints[0].y-1,
				  (unsigned char *)im->gdes[i].gd_set->ads_name,
				  graph_col[GRC_FONT].i);
		} else {
		    polyPoints[0].x = im->gdes[i].legloc.x;
		    polyPoints[0].y = im->gdes[i].legloc.y;
		    
		    gdImageString(gif, SmallFont,
				  polyPoints[0].x, 
				  polyPoints[0].y,
				  (unsigned char *)im->gdes[i].gd_set->ads_name,
				  graph_col[GRC_FONT].i);
		}
	}
    }
    
    
    /*
    gator(gif, (int) im->xgif-5, 5);
    */

}


static gdImagePtr
MkLineBrush(image_desc_t *im,long cosel, int width){
  gdImagePtr brush;
  int pen;
  int i, j;
  brush = gdImageCreate(width,width);

  gdImageColorTransparent(brush, 
			  gdImageColorAllocate(brush, 0, 0, 0));

  pen = gdImageColorAllocate(brush, 
			     im->gdes[cosel].col.red,
			     im->gdes[cosel].col.green,
			     im->gdes[cosel].col.blue);
    
  for (i = 0; i < width; i++) {
	  for (j = 0; j < width; j++) {
		if (0 != (width % 2) &&
			(i == 0 || i == (width-1)) &&
			(j == 0 || j == (width-1)) &&
			width > 1) {
			continue;
		}
		gdImageSetPixel(brush,i,j,pen);
	  }
  }
  return brush;
}

/* draw that picture thing ... */

static int
graph_paint(image_desc_t *im)
{
    int i;
    FILE  *fo;
    
    /* gif stuff */
    gdImagePtr	gif;

    double areazero = 0.0;
    gdPoint canvas[4], back[4];	 /* points for canvas*/

    /* get actual drawing data and find min and max values*/
    if(data_proc(im)==-1)
	return -1;

    if(!im->logarithmic){si_unit(im);}        /* identify si magnitude Kilo, Mega Giga ? */

    if(!im->rigid && ! im->logarithmic)
	expand_range(im);   /* make sure the upper and lower limit are
			   sensible values */

    /* init xtr and ytr */

    im->xgif = im->opts->go_width;
    im->ygif = im->opts->go_height;

    /*
     *  This is all voodoo.  I modified the constants from
     *  rrd to my taste.
     */
    if (im->tiny) {
	/*
	 *  No margins, no labels.
	 */
	im->xorigin = 0;
	im->xsize = im->xgif;
    } else {
        im->xorigin = 10 + 10 * SmallFont->w+SmallFont->h;
        im->xorigin = AD_GRAPH_LEFT_MARGIN_PIXELS +
		(AD_GRAPH_FORMATTED_YLABEL_WIDTH +
		 strlen(im->opts->go_ysuffix)) * SmallFont->w;

        im->xsize = im->xgif - AD_GRAPH_RIGHT_MARGIN_PIXELS - im->xorigin;
    }

    if (im->tiny) {
        im->yorigin = im->ygif;
        im->ysize = im->yorigin;
    } else {
        im->yorigin = im->ygif - AD_GRAPH_BOTTOM_MARGIN_PIXELS -
		SmallFont->h;

        im->ysize = im->yorigin - AD_GRAPH_TOP_MARGIN_PIXELS;
    }

    if(im->opts->go_title[0] != '\0' && !im->tiny)
	im->ysize -= (LargeFont->h+AD_GRAPH_TITLE_PADDING_PIXELS);

    /* determine where to place the legends onto the graphics.
       and set im->ygif to match space requirements for text */
    if(leg_place(im)==-1)
     return -1;

    xtr(im,0); 
    ytr(im,DNAN);

    gif=gdImageCreate(im->xgif,im->ygif);

    gdImageInterlace(gif, im->opts->go_interlaced);
    
    /* allocate colors for the screen elements */
    for(i=0;i<DIM(graph_col);i++)
	/* check for user override values */
	if(im->graph_col[i].red != -1)
	    graph_col[i].i = 
		gdImageColorAllocate( gif,
				      im->graph_col[i].red, 
				      im->graph_col[i].green, 
				      im->graph_col[i].blue);
	else
	    graph_col[i].i = 
		gdImageColorAllocate( gif,
				      graph_col[i].red, 
				      graph_col[i].green, 
				      graph_col[i].blue);
	
    
    /* allocate colors for the graph */
    for(i=0;i<im->gdes_c;i++)
	/* only for elements which have a color defined */
	if (im->gdes[i].col.red != -1)
	    im->gdes[i].col.i = 
		gdImageColorAllocate(gif,
				     im->gdes[i].col.red,
				     im->gdes[i].col.green,
				     im->gdes[i].col.blue);
    
    
    /* the actual graph is created by going through the individual
       graph elements and then drawing them */
    
    back[0].x = 0;
    back[0].y = 0;
    back[1].x = back[0].x+im->xgif;
    back[1].y = back[0].y;
    back[2].x = back[1].x;
    back[2].y = back[0].y+im->ygif;
    back[3].x = back[0].x;
    back[3].y = back[2].y;

    gdImageFilledPolygon(gif,back,4,graph_col[GRC_BACK].i);

    canvas[0].x = im->xorigin;
    canvas[0].y = im->yorigin;
    canvas[1].x = canvas[0].x+im->xsize;
    canvas[1].y = canvas[0].y;
    canvas[2].x = canvas[1].x;
    canvas[2].y = canvas[0].y-im->ysize;
    canvas[3].x = canvas[0].x;
    canvas[3].y = canvas[2].y;

    gdImageFilledPolygon(gif,canvas,4,graph_col[GRC_CANVAS].i);

    if (im->opts->go_minval > 0.0)
	areazero = im->opts->go_minval;
    if (im->opts->go_maxval < 0.0)
	areazero = im->opts->go_maxval;

    axis_paint(im,gif);

    grid_paint(im,gif);

    data_paint(im,gif);

#ifdef	AD_GRAPH_RULE_SUPPORT
    /* the RULES are the last thing to paint ... */
    for(i=0;i<im->gdes_c;i++){	
        
	switch(im->gdes[i].gf){
	case GF_HRULE:
	    if(im->gdes[i].yrule >= im->opts->go_minval
	       && im->gdes[i].yrule <= im->opts->go_maxval)
	      gdImageLine(gif,
			  im->xorigin,ytr(im,im->gdes[i].yrule),
			  im->xorigin+im->xsize,ytr(im,im->gdes[i].yrule),
			  im->gdes[i].col.i); 
	    break;
	case GF_VRULE:
	  if(im->gdes[i].xrule >= im->opts->go_start
	     && im->gdes[i].xrule <= im->opts->go_end)
	    gdImageLine(gif,
			xtr(im,im->gdes[i].xrule),im->yorigin,
			xtr(im,im->gdes[i].xrule),im->yorigin-im->ysize,
			im->gdes[i].col.i); 
	  break;
	default:
	    break;
	}
    }
#endif

    if (strcmp(im->opts->go_file,"-")==0) {
	fo = stdout;
    } else {
	if ((fo = fopen(im->opts->go_file,"wb")) == NULL) {
	    rrd_set_error("Openin %s for write: %s",im->opts->go_file, strerror(errno));
	    return (-1);
	}
    }
    switch (im->opts->go_imgformat) {
    case Format_Gif:
	if (im->opts->go_headers) {
		fprintf(fo, "Content-type: image/gif\n\n");
	}
	gdImageGif(gif, fo);    
	break;
    case Format_Png:
	if (im->opts->go_headers) {
		fprintf(fo, "Content-type: image/png\n\n");
	}
	gdImagePng(gif, fo);    
	break;
    }
    if (strcmp(im->opts->go_file,"-") != 0)
	fclose(fo);
    gdImageDestroy(gif);

    return 0;
}


void
ad_graph_init(ad_graph_opt_t * options)
{
	options->go_file[0] = '-';
	options->go_file[1] = '\0';
	options->go_width = 400;
	options->go_height = 100;
	options->go_ylegend[0] = 0;
	options->go_ysuffix[0] = 0;
	options->go_title[0] = 0;
	options->go_imgformat = Format_Gif;

	options->go_start = 0;
	options->go_end = 0;
	options->go_base = 1000;
	options->go_headers = 0;
	options->go_logarithmic = 0;
	options->go_interlaced = 0;
	options->go_rigid = 0;
	options->go_arrow = 0;
	options->go_hgrid = AD_GRID_MAJOR;
	options->go_vgrid = AD_GRID_NONE;
	options->go_legend = 0;
	options->go_extra_flags = 0;

	options->go_minval = DNAN;
	options->go_maxval = DNAN;    
	options->go_set_count = 0;
	options->go_sets = 0;
}

static void
gdes_init(image_desc_t * im, graph_desc_t * gdes, ad_set_t * set)
{
    gdes->gd_set = set;

    ad_graph_string_to_color(set->ads_color, &gdes->col);
    gdes->format[0] = 0;
}

int 
ad_graph(ad_graph_opt_t * opts)
{
    image_desc_t   im = {0};
    int            i;

    im.opts = opts;
    im.xlab_user.minsec = -1;
    im.xgif=opts->go_width;
    im.ygif=opts->go_height;
    if (im.xgif < AD_GRAPH_TINY_THRESHOLD ||
	im.ygif < AD_GRAPH_TINY_THRESHOLD) {
	im.tiny = 1;
    }
    else {
        im.tiny = 0;
    }

    if (im.ygif > AD_GRAPH_LARGE_THRESHOLD) {
	im.large = 1;
    }
    else {
	im.large = 0;
    }
    if (0 != im.large) {
	im.opts->go_legend = 1;
    }

    /*
     *  The bands look silly on small graphs.
     *  Use them only for large ones.
     */
    if (im.large) {
	im.opts->go_hgrid |= AD_GRID_BAND;
    }
    im.rigid = 0;
    im.ygridstep = DNAN;
    im.prt_c = 0;
    im.gdes_c = opts->go_set_count;
    im.gdes = calloc(im.gdes_c, sizeof(graph_desc_t));

    for ( i = 0; i < im.gdes_c; i++) {
	gdes_init(&im, &im.gdes[i], opts->go_sets[i]);
    }

    for(i=0;i<DIM(graph_col);i++)
	im.graph_col[i].red=-1;
    
    if (im.opts->go_logarithmic == 1 && (im.opts->go_minval <= 0 || isnan(im.opts->go_minval))){
	rrd_set_error("for a logarithmic yaxis you must specify a lower-limit > 0");	
	return -1;
    }

    if (im.opts->go_set_count == 0){
	rrd_set_error("can't make a graph without contents");
	return(-1); 
    }
    
	/* parse rest of arguments containing information on what to draw*/
    if (graph_paint(&im)==-1){
	return -1;
    }
    
    return 0;
}

static void
ad_graph_append_sample(ad_set_t * set, ad_time_t ts, ad_value_t val)
{
	ad_sample_t *	sample;

	sample = calloc(1, sizeof(ad_sample_t));
	sample->s_time = ts;
	sample->s_value = val;

	if (0 == set->ads_sample_allocated_count) {
		set->ads_samples = calloc(16 + 1, sizeof(ad_sample_t));
		set->ads_sample_allocated_count = 16;
	}
	else if (set->ads_sample_count >= set->ads_sample_allocated_count) {
		set->ads_sample_allocated_count *= 2;
		set->ads_samples = realloc(set->ads_samples,
			( set->ads_sample_allocated_count + 1 ) *
			sizeof(ad_sample_t));
	}

	set->ads_samples[ set->ads_sample_count ++ ] = sample;
	set->ads_samples[ set->ads_sample_count ] = 0;
}

#define AD_GRAPH_MAX_ARGV	16

static void
ad_graph_line_to_argv(char * line, int * argcp, char *** argvp)
{
	char * ptr;
	char * end;

	*argvp = calloc(AD_GRAPH_MAX_ARGV + 1, sizeof(char *));

	*argcp = 0;

	end = ptr = line;
	for ( *argcp = 0; *argcp < AD_GRAPH_MAX_ARGV ; ) {
		if (NULL == end || 0 == *end) {
			break;
		}
		while (isspace((int) *end)) {
			end++;
		}
		ptr = end;
		if ('"' == *ptr) {
			char * equote;

			ptr++;
			equote = strchr(ptr, '"');
			if (NULL == equote) {
				/*
				 *  Unterminated quote.
				 */
				end = NULL;
			}
			else {
				*equote = 0;
				end = equote + 1;
			}
		}
		else {
			end = strpbrk(ptr, " \t\n");
			if (NULL != end) {
				*end = 0;
				end++;
			}
		}
		(*argvp)[ (*argcp) ++ ] = ptr;
		(*argvp)[ (*argcp) ] = 0;
	}
	return;
}

static ad_graph_style_t
ad_graph_strtostyle(const char * string) {
    if (0 == strcasecmp(string, "LINE"))	{ return Graph_Line; }
    if (0 == strcasecmp(string, "BAR"))		{ return Graph_Bar; }
    if (0 == strcasecmp(string, "POINT"))	{ return Graph_Point; }
    return 0;
}

static ad_set_t *
ad_graph_new_set(char * line)
{
	ad_set_t *	set;
	int		argc;
	char **		argv;

	ad_graph_line_to_argv(line, &argc, &argv);

	set = calloc(1, sizeof(ad_set_t));

	optind = 1;
	while (1){
	static struct option long_options[] =
	{
	    {"width",      required_argument, 0,  'w'},
	    {"title",      required_argument, 0,  't'},
	    {"style",      required_argument, 0,  's'},
	    {"color",      required_argument, 0,  'c'},
	    {0,0,0,0}};
	int option_index = 0;
	int opt;

	opt = getopt_long(argc, argv, 
			  "w:t:s:c:",
			  long_options, &option_index);

	if (opt == EOF)
	    break;
	
	switch(opt) {
	case 'w':
	    set->ads_width = atol(optarg);
	    break;
	case 't':
	    strcpy(set->ads_name, optarg);
	    break;
	case 's':
	    set->ads_style = ad_graph_strtostyle(optarg);
	    break;
	case 'c':
	    strcpy(set->ads_color, optarg);
	    break;
	case '?':
            if (optopt != 0)
                rrd_set_error("unknown option '%c'", optopt);
            else
                rrd_set_error("unknown option '%s'",argv[optind-1]);
            return 0;
	}
    }

	return set;
}
static int
ad_graph_read(ad_graph_opt_t * opts, FILE * fp)
{
	char	buf [ AD_GRAPH_LINE_LEN + 1 ];
	time_t	date;
	char *	ptr;

	size_t		setcnt = 0;
	ad_set_t ** 	sets = 0;
	ad_set_t * 	set;

	time_t	mindate = time(0);
	time_t	maxdate = 0;

	size_t		column = 0;

	for ( ; ; ) {
		ptr = fgets(buf, AD_GRAPH_LINE_LEN, fp);
		if (NULL == ptr) {
			break;
		}

		if (0 == strncmp(buf, "#title ", 7)) {
			buf[ strlen(buf) - 1 ] = 0;
			strcpy(opts->go_title, &buf[7]);
			continue;
		}
		else if (0 == strncmp(buf, "#data-set ", 10)) {
			set = ad_graph_new_set(buf);

			if (0 != set) {
				if (0 == sets) {
					sets = calloc(setcnt + 1, 
						sizeof(ad_set_t *));
				}
				else {
					sets = realloc(sets,
						(setcnt + 1) *
						sizeof(ad_set_t *));
				}
				sets[setcnt] = set;
				setcnt++;
			}
			continue;
		}

		ptr = strtok(buf, " \t\n");

		if (NULL == ptr) {
			continue;
		}
		date = strtoul(ptr, NULL, 0);

		for ( column = 0; column < setcnt; column++ ) {
			ptr = strtok(NULL, " \t\n");

			if (NULL == ptr) {
				break;
			}
			if (0 == strcmp("-", ptr)) {
				continue;
			}

			ad_graph_append_sample(sets[column], 
					date, strtod(ptr, NULL));
			
			if (date < mindate) {
				mindate = date;
			}
			if (date > maxdate) {
				maxdate = date;
			}
		}
	}

	opts->go_sets = sets;
	opts->go_set_count = setcnt;

	if (0 == opts->go_start) {
		opts->go_start = mindate;
	}
	if (0 == opts->go_end) {
		opts->go_end = maxdate;
	}
	return 0;
}

static ad_graph_format_t if_conv(char *string){
    
    if (0 == strcmp(string, "GIF"))	{ return Format_Gif; }
    if (0 == strcmp(string, "PNG"))	{ return Format_Png; }

    return -1;
}

static int 
ad_graph_main(int argc, char **argv)
{
    long long_tmp;
    ad_graph_opt_t opts;
    char *	infile = 0;

    ad_graph_init(&opts);

    optind = 1;
    while (1){
	static struct option long_options[] =
	{
	    {"start",      required_argument, 0,  's'},
	    {"end",        required_argument, 0,  'e'},
	    {"width",      required_argument, 0,  'w'},
	    {"height",     required_argument, 0,  'h'},
	    {"upper-limit",required_argument, 0,  'u'},
	    {"lower-limit",required_argument, 0,  'l'},
	    {"rigid",      no_argument,       0,  'r'},
	    {"base",       required_argument, 0,  'b'},
	    {"title",      required_argument, 0,  't'},
	    {"imgformat",  required_argument, 0,  'a'},
	    {"ysuffix",	   required_argument, 0,  'Y'},
	    {"headers",	   no_argument,       0,  'H'},
	    {"outfile",	   required_argument, 0,  'o'},
	    {"infile",	   required_argument, 0,  'f'},
	    {0,0,0,0}};
	int option_index = 0;
	int opt;

	opt = getopt_long(argc, argv, 
			  "s:e:w:h:u:l:rb:t:a:Y:Ho:f:",
			  long_options, &option_index);

	if (opt == EOF)
	    break;
	
	switch(opt) {
	case 's':
	    opts.go_start = atol(optarg);
	    break;
	case 'e':
	    opts.go_end = atol(optarg);
	    break;
	case 'u':
	    opts.go_maxval = atof(optarg);
	    break;
	case 'l':
	    opts.go_minval = atof(optarg);
	    break;
	case 'b':
	    opts.go_base = atol(optarg);
	    if(opts.go_base != 1024 && opts.go_base != 1000 ){
		rrd_set_error("the only sensible value for base apart from 1000 is 1024");
		return -1;
	    }
	    break;
	case 'w':
	    long_tmp = atol(optarg);
	    if (long_tmp < 5) {
		rrd_set_error("width below 10 pixels");
		return -1;
	    }
	    opts.go_width = long_tmp;
	    break;
	case 'h':
	    long_tmp = atol(optarg);
	    if (long_tmp < 5) {
		rrd_set_error("height below 10 pixels");
		return -1;
	    }
	    opts.go_height = long_tmp;
	    break;
	case 'r':
	    opts.go_rigid = 1;
	    break;
    	case 'a':
	    if((opts.go_imgformat = if_conv(optarg)) == -1) {
		rrd_set_error("unsupported graphics format '%s'",optarg);
		return -1;
	    }
	    break;
	case 'f':
	    infile = optarg;
	    break;
	case 'o':
	    strcpy(opts.go_file, optarg);
	    break;
	case 't':
	    strncpy(opts.go_title,optarg,150);
	    opts.go_title[150]='\0';
	    break;

	case 'Y':
	    strncpy(opts.go_ysuffix, optarg, 15);
	    opts.go_ysuffix[15] = 0;
	    break;
	
	case 'H':
	    opts.go_headers = 1;
	    break;

	case '?':
            if (optopt != 0)
                rrd_set_error("unknown option '%c'", optopt);
            else
                rrd_set_error("unknown option '%s'",argv[optind-1]);
            return -1;
	}
    }

    if (0 == infile || 0 == strcmp("-", infile)) {
	ad_graph_read(&opts, stdin);
    }
    else {
	FILE *	fp;

	fp = fopen(infile, "r");
	if (NULL == fp) {
		rrd_set_error("can't open input file '%s'", infile);
		return -1;
	}
	ad_graph_read(&opts, fp);
	fclose(fp);
    }
    return ad_graph(&opts);
}

int
main(int argc, char ** argv)
{
	int rc;

	rc = ad_graph_main(argc, argv);

	if (0 != rc) {
		fprintf(stderr, "%s\n", rrd_get_error());
		exit(1);
	}
	exit(0);
}
