/*
 * $Id: //depot/prod/champagne/Rchampagne5.0.1x/src/rrdtool-1.0.28/src/ad_graph.h#1 $
 *
 * Copyright (c) 2003 Network Appliance, Inc.
 * All rights reserved.
 */

#ifndef _RRD_AD_GRAPH_H
#define _RRD_AD_GRAPH_H 1

/*------------------------------------------------------------------
 *
 *  ad_graph - graphing of arbitrary data
 *
 *  This is based largely on the RRD graphing implementation, with
 *  some changes:
 *
 *	- data is provided by the caller, not from an RRD (hence
 *	  the name)
 *
 *	- rather than a rigorous requirement that all the data
 *	  have a step size, this interface requires the caller
 *	  to supply the timestamp associated with each data
 *	  sample
 *
 *	- caller is responsible for all derived values (rather
 *	  than having this module implement the CDEF, etc)
 *
 *------------------------------------------------------------------*/

#include <stdio.h>
#include "rrd.h"
#include "rrd_tool.h"

/*
 *  Exported constants:
 */

#define	AD_LEGEND_LEN	2000

/*
 *  Exported type definitions:
 */

typedef	double	ad_value_t;
typedef time_t	ad_time_t;

typedef enum {
	/*
	 *  A line graph with the points connected.
	 */
	Graph_Line = 0,

	/*
	 *  Each data point is drawn as a vertical var.
	 */
	Graph_Bar,

	/*
	 *  Only draw the data points.
	 */
	Graph_Point,
} ad_graph_style_t;

typedef struct {
	ad_time_t	s_time;
	ad_value_t	s_value;
} ad_sample_t;

typedef struct {
	char		ads_name[100];
	int		ads_width;
	ad_graph_style_t ads_style;
	char 		ads_color[16];
	size_t		ads_sample_count;
	ad_sample_t **	ads_samples;

	size_t		ads_sample_allocated_count;
} ad_set_t;

typedef enum {
	Format_Gif,
	Format_Png

} ad_graph_format_t ;

#define	AD_GRID_NONE	(0)
#define	AD_GRID_MAJOR	(1)
#define	AD_GRID_MINOR	(2)
#define	AD_GRID_BAND	(4)	/* can be OR-ed with MAJOR or MINOR */

typedef struct {
    char           go_file[MAXPATH]; /* filename for graphic */
    long           go_width;
    long	   go_height;    /* graph area size in pixels */
    char           go_ylegend[200];   /* legend along the yaxis */
    char           go_ysuffix[16];   /* suffix for y axix labels */
    char           go_title[200];     /* title for graph */
    ad_graph_format_t go_imgformat;

    ad_time_t      go_start;
    ad_time_t	   go_end;
    long	   go_base;
    int		   go_headers;	      /* print http headers */
    int            go_logarithmic;    /* scale the yaxis logarithmic */
    int            go_interlaced;     /* will the graph be interlaced? */
    int		   go_rigid;
    int		   go_arrow;
    int		   go_hgrid;
    int		   go_vgrid;
    int		   go_legend;
    int            go_extra_flags;    /* flags for boolean options */

    ad_value_t     go_minval;
    ad_value_t	   go_maxval;
    size_t	   go_set_count;
    ad_set_t **	   go_sets;

} ad_graph_opt_t;

/*
 *  Exported variables:
 */

/*
 *  Exported functions:
 */

extern void ad_graph_init(ad_graph_opt_t * options);

extern int ad_graph(ad_graph_opt_t * options);

#endif
